import argparse
import sys
import pysam
import re
import math

from marker_alignments.store import AlignmentStore
from marker_alignments.write import write, output_type_options
from marker_alignments.mcl import clusters 
from marker_alignments.refdb_pattern import taxon_and_marker_patterns

from marker_alignments.pysam2 import compute_contribution_to_marker_coverage, compute_alignment_identity

def next_g(search):
    return next(g for g in search.groups() if g is not None)

def taxon_and_marker(reference_name, pattern_taxon, pattern_marker, marker_to_taxon):

    taxon = marker_to_taxon[reference_name] if reference_name in marker_to_taxon else None
    taxon_search = pattern_taxon.search(reference_name)
    if taxon_search and taxon:
        taxon = "|".join([taxon, next_g(taxon_search)])
    elif taxon_search:
        taxon = next_g(taxon_search)
    elif taxon:
        taxon = taxon
    else:
        taxon = None

    marker_search = pattern_marker.search(reference_name)
    if marker_search:
        marker=next_g(marker_search)
    elif taxon_search:
        # Regexes seem to work, just not this one 
        marker = None
    else:
        marker = reference_name
    return (taxon, marker)

def read_alignments(alignment_file, sqlite_db_path, pattern_taxon, pattern_marker, marker_to_taxon, min_mapq, min_query_length, min_match_identity):

    alignment_store = AlignmentStore(db_path=sqlite_db_path)
    alignment_store.start_bulk_write()

    for read in alignment_file.fetch():
        identity = compute_alignment_identity(read)
        if read.mapq < min_mapq:
            continue
        if read.infer_query_length() < min_query_length:
            continue
        if identity < min_match_identity:
            continue
        if not read.reference_name:
            raise ValueError("Read missing reference name: " + str(read))
        (taxon, marker) = taxon_and_marker(read.reference_name, pattern_taxon, pattern_marker, marker_to_taxon)
        if not taxon:
            raise ValueError("Could not find taxon in reference name: " + read.reference_name)
        if not marker:
            raise ValueError("Could not find marker in reference name: " + read.reference_name)

        alignment_store.add_alignment(
          taxon = taxon,
          marker = marker,
          query = read.query_name,
          identity = identity,
          coverage = compute_contribution_to_marker_coverage(alignment_file, read),
        )

    alignment_store.end_bulk_write()
    return alignment_store

def read_marker_to_taxon(path):
    result = {}
    with open(path, 'r') as f:
        for line in f:
            (marker, taxon) = line.rstrip().split("\t")
            result[marker] = taxon
    return result

def parse_arguments(argv=sys.argv[1:]):
    parser = argparse.ArgumentParser(
      description="summarize_marker_alignments - process and summarise alignments of metagenomic sequencing reads to reference databases of marker genes",
      formatter_class = argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument("--input", type=str, action="store", dest="input_alignment_file", help = "Input SAM/BAM", required=True)
    parser.add_argument("--sqlite-db-path", type=str, action="store", dest="sqlite_db_path", help = "Store a sqlite database under this path instead of in memory", default=None)
    parser.add_argument("--refdb-format", type=str, action="store", dest="refdb_format", help = "Reference database used for alignment, required for parsing reference names. Supported values: eukprot, chocophlan, generic, no-split (no split into marker and taxon)", default="generic")
    parser.add_argument("--refdb-regex-taxon", type=str, action="store", dest="refdb_regex_taxon", help = "Regex to read taxon name from reference name")
    parser.add_argument("--refdb-regex-marker", type=str, action="store", dest="refdb_regex_marker", help = "Regex to read marker name from reference name")
    parser.add_argument("--refdb-marker-to-taxon-path", type=str, action="store", dest="refdb_marker_to_taxon_path", help = "Lookup file, two columns - marker name, taxon name")
    parser.add_argument("--num-reads", type=int, action="store", dest="num_reads", help = "Total number of reads (required for CPM output)")
    parser.add_argument("--output-type", type=str, action="store", dest="output_type", help = "output type: "+", ".join(output_type_options), default = "marker_coverage")
    parser.add_argument("--output", type=str, action="store", dest="output_path", help = "output path", required=True)
    parser.add_argument("--min-read-mapq", type=int, action="store", dest="min_read_mapq", help = "when reading the input, skip alignments with MAPQ < min-read-mapq", default=0)
    parser.add_argument("--min-read-query-length", type=int, action="store", dest="min_read_query_length", help = "when reading the input, skip alignments shorter than min-read-query-length", default=0)
    parser.add_argument("--min-read-match-identity", type=float, action="store", dest="min_read_match_identity", help = "when reading the input, skip alignments where the proportion of matching bases in the alignment is less than min-read-match-identity", default=0)
    parser.add_argument("--min-taxon-num-markers", type=int, action="store", dest="min_taxon_num_markers", help = "Only keep taxa with at least min-taxon-num-markers markers")
    parser.add_argument("--min-taxon-num-reads", type=int, action="store", dest="min_taxon_num_reads", help = "Only keep taxa with at least min-taxon-num-reads reads")
    parser.add_argument("--min-taxon-num-alignments", type=int, action="store", dest="min_taxon_num_alignments", help = "Only keep taxa with at least min-taxon-num-alignments alignments")
    parser.add_argument("--min-taxon-fraction-primary-matches", type=float, action="store", dest="min_taxon_fraction_primary_matches", help = "Only keep taxa where no more than min-taxon-fraction-primary-matches fraction of alignments is inferior / secondary")
    parser.add_argument("--min-taxon-better-marker-cluster-averages-ratio", type=float, action="store", dest="min_taxon_better_cluster_averages_ratio", help = "Only keep taxa where the ratio between markers which have at least average match identity relative to their clusters and markers with identity below average is at least min-taxon-better-cluster-averages-ratio")

    parser.add_argument("--threshold-avg-match-identity-to-call-known-taxon", type=float, action="store", dest="threshold_identity_to_call_taxon", help = "Threshold on average match identity to return taxon in reference")
    parser.add_argument("--threshold-num-reads-to-call-unknown-taxon", type=int, action="store", dest="threshold_num_reads_to_call_unknown_taxon", help = "To positively identify an unknown taxon (fits all criteria except match identity) expect this many reads from a taxon cluster")
    parser.add_argument("--threshold-num-markers-to-call-unknown-taxon", type=int, action="store", dest="threshold_num_markers_to_call_unknown_taxon", help = "To positively identify an unknown taxon (fits all criteria except match identity) expect this many markers from a taxon cluster")
    parser.add_argument("--threshold-num-taxa-to-call-unknown-taxon", type=int, action="store", dest="threshold_num_taxa_to_call_unknown_taxon", help = "To positively identify an unknown taxon (fits all criteria except match identity) expect this many taxa from a taxon cluster")
    result = parser.parse_args(argv)
    return result

def main(argv=sys.argv[1:]):
    options=parse_arguments(argv)

    if options.min_read_mapq and (options.min_taxon_fraction_primary_matches or options.min_taxon_better_cluster_averages_ratio) :
        raise ValueError("It us unwise to combine --min-read-mapq and filters that rely on secondary matches!")

    if options.output_type not in output_type_options:
        raise ValueError("Unknown output type: " + options.output_type + ". Please choose one of the following: " + ", ".join(output_type_options))

    if options.refdb_format:
        (tp, mp) = taxon_and_marker_patterns(options.refdb_format)
        if not (tp and mp):
            raise ValueError("Unknown refdb format: " + options.refdb_format)
        options.refdb_regex_taxon=tp
        options.refdb_regex_marker=mp

    if not (options.refdb_regex_taxon and options.refdb_regex_marker):
        raise ValueError("Please provide either refdb format, or taxon + marker regexes")

    if options.output_type in ["marker_all","marker_cpm", "taxon_all", "taxon_cpm"] and not options.num_reads:
        raise ValueError("--num-reads required for calculating " + options.output_type)

    alignment_store = read_alignments(
      alignment_file = pysam.AlignmentFile(options.input_alignment_file),
      sqlite_db_path = options.sqlite_db_path,
      marker_to_taxon = read_marker_to_taxon(options.refdb_marker_to_taxon_path) if options.refdb_marker_to_taxon_path else {},
      pattern_taxon = re.compile(options.refdb_regex_taxon),
      pattern_marker = re.compile(options.refdb_regex_marker),
      min_mapq = options.min_read_mapq,
      min_query_length = options.min_read_query_length,
      min_match_identity = options.min_read_match_identity,
    )

    alignment_store.cluster_markers_by_matches()

    if options.min_taxon_better_cluster_averages_ratio:
        alignment_store.modify_table_filter_taxa_on_cluster_averages(min_better_cluster_averages_ratio = options.min_taxon_better_cluster_averages_ratio)

    if options.min_taxon_fraction_primary_matches:
        alignment_store.modify_table_filter_taxa_on_multiple_matches(min_fraction_primary_matches = options.min_taxon_fraction_primary_matches)

    if options.min_taxon_num_markers or options.min_taxon_num_reads or options.min_taxon_num_alignments:
        alignment_store.modify_table_filter_taxa_on_num_markers_reads_and_alignments(min_num_markers = options.min_taxon_num_markers or 0, min_num_reads = options.min_taxon_num_reads or 0, min_num_alignments = options.min_taxon_num_alignments or 0)


    alignment_store.cluster_taxa_by_matches()

    if options.threshold_identity_to_call_taxon or options.threshold_num_reads_to_call_unknown_taxon or options.threshold_num_markers_to_call_unknown_taxon or options.threshold_num_taxa_to_call_unknown_taxon:
        alignment_store.modify_table_transform_taxa_on_thresholds_and_clusters(
                threshold_identity = options.threshold_identity_to_call_taxon or 0,
                min_num_taxa_below_identity = options.threshold_num_taxa_to_call_unknown_taxon or 0,
                min_num_markers_below_identity = options.threshold_num_markers_to_call_unknown_taxon or 0,
                min_num_reads_below_identity = options.threshold_num_reads_to_call_unknown_taxon or 0
        )

    write(alignment_store, options.output_type, options.output_path, options.num_reads)



# todo: SS SG MG collapse

eukprot_refdb_regex_taxon = "^[a-z]+-(.*_[a-z-]+(?:_.*)?)-[0-9]+at2759-[A-Z]\d$|^[a-z]+-(.*)...Collapse.*$"
eukprot_refdb_regex_marker = "^[a-z]+-.*_[a-z-]+(?:_.*)?-([0-9]+at2759-[A-Z]\d)$|^[a-z]+-.*(..Collapse.*)$"

chocophlan_refdb_regex_taxon="s__(.*?)\\|"
chocophlan_refdb_regex_marker="(UniRef90[^|]*)"

no_split_regex_taxon="(^)"
no_split_regex_marker="(.*)"

def taxon_and_marker_patterns(refdb_format):
    if refdb_format == "eukprot":
        return (eukprot_refdb_regex_taxon, eukprot_refdb_regex_marker)
    elif refdb_format == "chocophlan":
        return (chocophlan_refdb_regex_taxon, chocophlan_refdb_regex_marker)
    elif refdb_format == "no-split":
        return (no_split_regex_taxon, no_split_regex_marker)
    elif refdb_format == "generic":
        return (
           "|".join([eukprot_refdb_regex_taxon, chocophlan_refdb_regex_taxon, "(^[^:|]*)[:|]", no_split_regex_taxon]),
           "|".join([eukprot_refdb_regex_marker, chocophlan_refdb_regex_marker, "[:|]?([^:|]*)$"]),
        )
    else:
        return (None, None)






# populated when the module loads
num_reads_arg_count_in_sql = {}
sqls = {}

# when splitting read stats by query, do it proportionally to the second power of match identity
# if there are multiple matches in a query + taxon + marker, return identity as max and coverage as weighted average
marker_query_template='''
  select taxon, marker, {} from (
    select
      a.query,
      a.taxon,
      a.marker,
      max(a.identity) as identity,
      {}
    from
      alignment a join (
      select query, sum(identity * identity) as total_weight_for_query
        from alignment group by query
      ) as m
    where a.query = m.query
    group by a.taxon, a.marker, a.query
  ) group by taxon, marker
'''
s_cov="sum(coverage) as marker_coverage"
p_cov="sum(a.coverage * a.identity * a.identity) / (m.total_weight_for_query) as coverage"
s_mrc="sum(alignment_count) as marker_alignment_count, sum(weight_fraction) as marker_read_count, avg(identity) as marker_avg_identity"
p_wf="sum(a.identity * a.identity) / (m.total_weight_for_query) as weight_fraction, count(*) as alignment_count"
s_cpm="sum(coverage) / (?) * 1000000 as marker_cpm"

sqls['marker_coverage'] = marker_query_template.format(s_cov, p_cov)
sqls['marker_read_count'] = marker_query_template.format(s_mrc, p_wf)
sqls['marker_cpm'] = marker_query_template.format(s_cpm, p_cov)
num_reads_arg_count_in_sql['marker_cpm'] = 1
sqls['marker_all'] = marker_query_template.format(", ".join([s_cov, s_cpm, s_mrc]), ", ".join([p_cov, p_wf]))
num_reads_arg_count_in_sql['marker_all'] = 1



taxon_query_template='''
  select taxon, {}
  from ({})
  group by taxon
'''

a_cov = "avg(marker_coverage) as coverage"
a_tnm = "sum(marker_read_count) as taxon_num_reads, sum(marker_alignment_count) as taxon_num_alignments, count(marker) as taxon_num_markers, max(marker_read_count) as taxon_max_reads_in_marker"
a_cpm = "avg(marker_coverage) / (?) * 1000000 as cpm"
sqls['taxon_coverage'] = taxon_query_template.format(a_cov, sqls['marker_coverage'])
sqls['taxon_read_and_marker_count'] = taxon_query_template.format(a_tnm, sqls['marker_read_count'])
sqls['taxon_cpm'] = taxon_query_template.format(a_cpm, sqls['marker_coverage'])
num_reads_arg_count_in_sql['taxon_cpm'] = 1
sqls['taxon_all'] = taxon_query_template.format(", ".join([a_cov, a_cpm, a_tnm]), sqls['marker_all'])
num_reads_arg_count_in_sql['taxon_all'] = 2


sqls['pairs_of_taxa_shared_queries'] = '''
select at as taxon,
       bt as related_taxon,
       sum(at_higher_identity_than_bt) as num_queries_where_taxon_higher_identity,
       sum(at_equal_or_higher_identity_than_bt) as num_queries_where_taxon_at_least_equal_identity,
       count(at_higher_identity_than_bt) as num_queries_shared
from   (select a.taxon at,
               b.taxon bt,
               max(a.identity) - max(b.identity) > 1e-6 as at_higher_identity_than_bt,
               max(a.identity) - max(b.identity) > -1e-6 as at_equal_or_higher_identity_than_bt
        from   alignment a,
               alignment b
        where  a.query = b.query
        and a.taxon != b.taxon
        group by a.query, at, bt
               )
group  by at,
          bt
order  by at
'''

sqls['taxa_in_marker_clusters']= '''select taxon,
  count(*) as taxon_num_marker_clusters,
  sum(higher_identity) as taxon_num_marker_clusters_at_least_cluster_average,
  sum(top_identity) as taxon_num_marker_clusters_best_in_cluster,
  sum(is_unique) as taxon_num_marker_clusters_unique_in_cluster
from (
  select t1.*,
    t2.avg_cluster_identity,
    t2.num_taxa == 1 as is_unique,
    t2.avg_cluster_identity - avg_identity < 1e-6 as higher_identity,
    t3.max_cluster_identity - avg_identity < 1e-6 as top_identity
    from (
      select id, mc.taxon, mc.marker, count(distinct query) as num_matches, avg(identity) as avg_identity
        from marker_cluster mc, alignment a
        where mc.taxon = a.taxon and mc.marker = a.marker
        group by id, mc.taxon, mc.marker
    ) t1, (
    select id, avg(a.identity) as avg_cluster_identity, count(distinct mc.taxon) as num_taxa
        from marker_cluster mc, alignment a
        where mc.taxon = a.taxon and mc.marker = a.marker
        group by id
    ) t2, (
    select id, max(avg_identity) as max_cluster_identity
    from (
          select id, mc.taxon, mc.marker, avg(identity) as avg_identity
            from marker_cluster mc, alignment a
            where mc.taxon = a.taxon and mc.marker = a.marker
            group by id, mc.taxon, mc.marker
        )
        group by id
    ) t3
    where t1.id = t2.id and t1.id = t3.id
   )
group by taxon
order by -taxon_num_marker_clusters, -taxon_num_marker_clusters_unique_in_cluster, -taxon_num_marker_clusters_best_in_cluster
'''
output_type_options = [k for k in sqls]

def get_output(alignment_store, output_type, num_reads):
    args = [sqls[output_type]]
    if output_type in num_reads_arg_count_in_sql:
        args.append([num_reads] * num_reads_arg_count_in_sql[output_type])
    return alignment_store.report(*args)

field_formats = {
  "taxon" : "",
  "marker": "",
  "marker_cpm": ":.6f",
  "marker_coverage": ":.6f",
  "marker_read_count": ":.2f",
  "marker_alignment_count": ":d",
  "marker_avg_identity": ":.6f",
  "cpm" : ":.6f",
  "coverage" : ":.6f",
  "taxon_num_alignments": ":d",
  "taxon_num_reads": ":.6f",
  "taxon_num_markers": ":d",
  "taxon_max_reads_in_marker": ":.6f",
  "related_taxon": "",
  "num_queries_where_taxon_higher_identity": ":d",
  "num_queries_where_taxon_at_least_equal_identity": ":d",
  "num_queries_shared":":d",
  "taxon_num_marker_clusters": ":d",
  "taxon_num_marker_clusters_at_least_cluster_average": ":d",
  "taxon_num_marker_clusters_best_in_cluster": ":d",
  "taxon_num_marker_clusters_unique_in_cluster": ":d",
}
def write(alignment_store, output_type, output_path, num_reads):
    header, lines = get_output(alignment_store, output_type, num_reads)
    formatter="\t".join(['{' + field_formats[field] +'}' for field in header]) + "\n"
    with open(output_path, 'w') as f:
        f.write("\t".join(header) + "\n")
        for line in lines:
            f.write(formatter.format(*line))



import pysam
import itertools

# longer genes have more reads aligning to them
# so instead of counting reads, estimate a number of copies of each marker
# this should be proportional to species abundance and agree between markers
# ( and also proportional to sequencing depth - we correct that later if calculating CPM output)
def compute_contribution_to_marker_coverage(alignment_file, sam_record):
    marker_length = alignment_file.get_reference_length(sam_record.reference_name)

    return round(1.0 * sam_record.infer_query_length() / marker_length, 6)

# adapted from https://github.com/mortazavilab/TALON/blob/master/src/talon/transcript_utils.py
def compute_alignment_identity(sam_record):
    """ This function computes what fraction of the read matches the reference
        genome."""

    read_ID = sam_record.query_name
    try:
        MD_tag = sam_record.get_tag('MD')
    except KeyError:
        raise ValueError("SAM transcript %s lacks an MD tag" % read_ID)

    total_bases = sam_record.infer_query_length()
    matches = 0.0
    ops, counts = splitMD(MD_tag)
    for op,ct in zip(ops, counts):
        if op == "M":
            matches += ct
        if op == "D":
            total_bases += ct

    return round(1.0 * matches/total_bases, 6)

def splitMD(MD):
        """ Takes MD tag and splits into two lists:
            one with capital letters (match operators), and one with
            the number of bases that each operation applies to. """

        operations = []

        # Split MD string where type changes.
        # Digits are separated from base changes.
        # Deletions (with ^) are captured together.
        counts = ["".join(x) for _, x in itertools.groupby(MD, key=str.isdigit)]

        # Get operations
        for i in range(0,len(counts)):
            curr = counts[i]
            try:
                counts[i] = int(curr)
                operations.append("M")
            except ValueError:
                # Handle deletion
                if curr.startswith("^"):
                    operations.append("D")
                    counts[i] = len(counts[i]) - 1
                else:
                    operations.append("X")
                    counts[i] = len(counts[i])

        return operations, counts



import setuptools

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setuptools.setup(name='marker_alignments',
    version='0.4.1',
    description="Process and summarise alignments of metagenomic sequencing reads to reference databases of marker genes",
    long_description = long_description,
    long_description_content_type="text/markdown",
    url='http://github.com/wbazant/marker_alignments',
    author='wbazant',
    author_email='wojciech.bazant@gmail.com',
    license="MIT",
    entry_points={"console_scripts": [
        "marker_alignments = marker_alignments.main:main",
    ]},
    install_requires=["pysam", "scipy", "numpy", "sklearn"],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
)



import sqlite3

from marker_alignments.mcl import clusters

class SqliteStore:
    def __init__(self, db_path = None):
        self.__db_path = db_path
        self.__conn = None
        self.__is_within_transaction = False
        self.__stateful_ops_in_bulk_write = None

    def connect(self):
        if self.__conn:
            return
        self.__conn = sqlite3.connect(self.__db_path or ":memory:", isolation_level=None)

    def do(self, *args):
        self.__conn.execute(*args)
        if self.__is_within_transaction:
            self.__stateful_ops_in_bulk_write +=1
            if self.__stateful_ops_in_bulk_write % 100000 == 0:
                self.__conn.execute("commit transaction")
                self.__conn.execute("begin transaction")

    def query(self, *args):
        return self.__conn.execute(*args)


    def start_bulk_write(self):
        self.__is_within_transaction = True
        self.__stateful_ops_in_bulk_write = 0
        self.__conn.execute("begin transaction")

    def end_bulk_write(self):
        self.__conn.execute("commit transaction")
        self.__is_within_transaction = False
        self.__stateful_ops_in_bulk_write = None


filter_taxa_on_multiple_matches_query = '''
  select a.* from alignment a,
  (
    select taxon,
       count(*) as num_matches,
       sum(is_unique) as num_unique_matches,
       sum(is_best) as num_best_matches,
       sum(is_inferior) as num_inferior_matches
    from   (select a.taxon,
      s.num_taxa == 1 as is_unique,
      s.num_taxa > 1 and s.top_identity - max(a.identity) < 1e-6 as is_best,
      s.num_taxa > 1 and s.top_identity - max(a.identity) > 1e-6 as is_inferior
      from   alignment a,
           (select query,
               Max(identity) as top_identity,
               count(distinct taxon) as num_taxa
          from   alignment
          group  by query) s
      where  a.query = s.query
      group by a.query, a.taxon
      )
    group  by taxon
  ) t
  where a.taxon = t.taxon and (t.num_unique_matches + t.num_best_matches ) >= (?) * t.num_matches
'''

# markers with only inferior alignments don't count
filter_taxa_on_num_markers_reads_and_alinments_query = '''
  select a.* from alignment a,
  (
    select taxon,
    count(distinct marker) as num_markers,
    count(distinct query) as num_reads
    from   (
      select a.taxon, a.marker, a.query
      from   alignment a,
           (select query,
               Max(identity) as top_identity,
               count(distinct taxon) as num_taxa
          from   alignment
          group  by query) s
      where  a.query = s.query
      group by a.query, a.taxon, a.marker
      having s.top_identity - max(a.identity) < 1e-6
      )
    group  by taxon
  ) t,
  (
    select taxon,
    count(*) as num_alignments
    from alignment
    group by taxon
  ) t2
  where a.taxon = t.taxon and a.taxon = t2.taxon and t.num_markers >= (?) and t.num_reads >= (?) and t2.num_alignments >= (?)
'''

filter_taxa_on_avg_identity_query = '''
  select a.* from alignment a,
  (
    select taxon,
      avg(top_identity) as avg_identity
    from   (select a.taxon,
            max(a.identity) as top_identity
            from alignment a
            group by a.query, a.taxon
            )
    group  by taxon
  ) t
  where a.taxon = t.taxon and t.avg_identity >= (?)
'''

filter_taxa_on_cluster_averages_query = '''
  select a.* from alignment a,
  (
    select taxon,
      sum(higher_identity) as num_markers_at_least_cluster_average,
      sum(lower_identity) as num_markers_below_cluster_average
    from (
  select t1.*,
        t2.avg_cluster_identity,
        t2.num_taxa,
        t2.avg_cluster_identity - avg_identity < 1e-6 as higher_identity,
        t2.avg_cluster_identity - avg_identity >= 1e-6 as lower_identity
        from (
          select id, mc.taxon, mc.marker, count(distinct query) as num_matches, avg(identity) as avg_identity
            from marker_cluster mc, alignment a
            where mc.taxon = a.taxon and mc.marker = a.marker
            group by id, mc.taxon, mc.marker
        ) t1, (
        select id, avg(identity) as avg_cluster_identity, count(distinct mc.taxon) as num_taxa
            from marker_cluster mc, alignment a
            where mc.taxon = a.taxon and mc.marker = a.marker
            group by id
        ) t2
        where t1.id = t2.id
       )
    group by taxon
  ) t
  where a.taxon = t.taxon and num_markers_at_least_cluster_average >= (?) * num_markers_below_cluster_average
'''

counts_of_common_matches_in_markers_query = '''
select 
       a.taxon at,
       a.marker am,
       b.taxon bt,
       b.marker bm,
       count(distinct a.query)
from   alignment a,
       alignment b
where  a.query = b.query
group by at, bt, am, bm;
'''

counts_of_common_matches_in_taxa_query = '''
select aa.at, aa.bt, cast (sum_shared  as real) / aaa.num_queries from
(
    select at, bt, count(*) as sum_shared
    from (
      select
           a.taxon at,
           b.taxon bt,
           a.query
      from   alignment a,
           alignment b
      where  a.query = b.query
      group by at, bt, a.query
    ) group by at, bt
) aa,
(
  select taxon, count(distinct query) as num_queries from alignment
  group by taxon
) aaa
where aa.at = aaa.taxon
'''


transform_taxa_on_thresholds_and_clusters_query = '''
select t.mapped_taxon as taxon, a.marker, a.query, a.identity, a.coverage from alignment a,
(
    select tc.taxon as original_taxon, tc.taxon as mapped_taxon
    from taxon_cluster tc, alignment al
    where tc.taxon = al.taxon 
    group by tc.id, tc.taxon
    having avg(al.identity) >= (?)

    union

    select 
      tc.taxon as original_taxon,
      m.mapped_taxon
    from taxon_cluster tc,
    (
      select id, '?' || group_concat(taxon) as mapped_taxon
      from (
        select tc.id,
          tc.taxon,
          count(distinct al.marker) as num_markers,
          count(distinct al.query) as num_reads,
          avg(al.identity) >= (?) as is_above_threshold
        from taxon_cluster tc, alignment al
        where tc.taxon = al.taxon 
        group by tc.id, tc.taxon
      ) group by id
      having
      (?) > 0 and count(distinct taxon) >= (?) and sum(num_markers) >= (?) and sum(num_reads) >= (?) and sum(is_above_threshold) == 0
    ) m
    where tc.id = m.id
) t
where a.taxon = t.original_taxon
'''
class AlignmentStore(SqliteStore):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.connect()
        self.do('''create table alignment (
              taxon text not null,
              marker text not null,
              query text not null,
              identity real not null,
              coverage real not null
            );''')
        self.do('''
            create table marker_cluster (
              id number not null,
              taxon text not null,
              marker text not null
            );''')
        self.do('''
            create table taxon_cluster (
              id number not null,
              taxon text not null
            );''')
        
    def add_alignment(self, taxon, marker, query, identity, coverage):
        self.do('insert into alignment (taxon, marker, query, identity, coverage) values (?,?,?,?,?)', [ taxon, marker, query, identity, coverage])


    def _modify_table(self, op, select_query, *args):
        self.query("begin transaction")
        self.query("create table new as " + select_query, *args)
        self.query("alter table alignment rename to alignment_pre_filter_on_" + op)
        self.query("alter table new rename to alignment")
        self.query("commit transaction")

    def modify_table_filter_taxa_on_multiple_matches(self, min_fraction_primary_matches):
        self._modify_table('multiple_matches', filter_taxa_on_multiple_matches_query, [min_fraction_primary_matches])

    def modify_table_filter_taxa_on_num_markers_reads_and_alignments(self, min_num_markers, min_num_reads, min_num_alignments):
        self._modify_table('num_markers', filter_taxa_on_num_markers_reads_and_alinments_query, [min_num_markers, min_num_reads, min_num_alignments])

    def modify_table_filter_taxa_on_avg_identity(self, min_avg_identity):
        self._modify_table('avg_identity', filter_taxa_on_avg_identity_query, [min_avg_identity])

    def modify_table_filter_taxa_on_cluster_averages(self, min_better_cluster_averages_ratio):
        self._modify_table('cluster_averages', filter_taxa_on_cluster_averages_query, [min_better_cluster_averages_ratio])

    def modify_table_transform_taxa_on_thresholds_and_clusters(self, threshold_identity, min_num_taxa_below_identity, min_num_markers_below_identity, min_num_reads_below_identity):
        try_return_unknown_taxa = 1 if min_num_taxa_below_identity or min_num_markers_below_identity or min_num_reads_below_identity else 0
        self._modify_table('thresholds_and_clusters', transform_taxa_on_thresholds_and_clusters_query,
	  [threshold_identity, threshold_identity, try_return_unknown_taxa, min_num_taxa_below_identity, min_num_markers_below_identity, min_num_reads_below_identity])

    def report(self, *args):
        cursor = self.query(*args)
        # https://stackoverflow.com/a/7831685
        description = next(zip(*cursor.description))
        # let the client iterate through the store
        return (description, cursor)

    def as_marker_coverage(self):
        return self.query(marker_coverage_query)

    def as_marker_read_count(self):
        return self.query(marker_read_count_query)

    def as_marker_cpm(self, total_reads):
        return self.query(marker_cpm_query, [total_reads])

    def as_marker_all(self, total_reads):
        return self.query(marker_all_query, [total_reads])

    def as_taxon_coverage(self):
        return self.query(taxon_coverage_query)

    def as_taxon_read_and_marker_count(self):
        return self.query(taxon_read_and_marker_count_query)

    def as_taxon_cpm(self,total_reads):
        return self.query(taxon_cpm_query, [total_reads])

    def as_taxon_all(self,total_reads):
        return self.query(taxon_all_query, [total_reads, total_reads])

    def cluster_markers_by_matches(self):
        triples = [(at + "\t" + am, bt + "\t" + bm, v)  for at, am, bt, bm, v in self.query(counts_of_common_matches_in_markers_query)]

        self._store_marker_clusters(clusters(triples))

    def _store_marker_clusters(self, clusters):

        self.start_bulk_write()
        for ix in range(0, len(clusters)):
            cluster = clusters[ix]
            cluster_id = ix + 1
            for x in cluster:
                taxon, marker = x.split("\t")
                self.do('insert into marker_cluster (id, taxon, marker) values (?,?,?)', [ cluster_id, taxon, marker])
        self.end_bulk_write()

    def cluster_taxa_by_matches(self):
        triples = [(at, bt, v)  for at, bt, v in self.query(counts_of_common_matches_in_taxa_query)]

        self._store_taxon_clusters(clusters(triples))

    def _store_taxon_clusters(self, clusters):

        self.start_bulk_write()
        for ix in range(0, len(clusters)):
            cluster = clusters[ix]
            cluster_id = ix + 1
            for taxon in cluster:
                self.do('insert into taxon_cluster (id, taxon) values (?,?)', [ cluster_id, taxon])
        self.end_bulk_write()
